# -*- coding: utf-8 -*-

import random
import numpy as np
import matplotlib.pyplot as plt
import time
import math

#set line width
plt.rcParams['lines.linewidth'] = 4
#set font size for titles 
plt.rcParams['axes.titlesize'] = 16
#set font size for labels on axes
plt.rcParams['axes.labelsize'] = 16
#set size of num_bers on x-axis
plt.rcParams['xtick.labelsize'] = 16
#set size of num_bers on y-axis
plt.rcParams['ytick.labelsize'] = 16
#set size of ticks on x-axis
plt.rcParams['xtick.major.size'] = 7
#set size of ticks on y-axis
plt.rcParams['ytick.major.size'] = 7
#set size of markers, e.g., circles representing points
plt.rcParams['lines.markersize'] = 10
#set num_ber of _times marker is shown when displaying legend
plt.rcParams['legend.numpoints'] = 1
#Set size of type in legend
plt.rcParams['legend.fontsize'] = 14


##################
## Example: Constant Complexity
##################
def add(x, y):
    return x+y


##################
## Example: Linear Complexity
##################
def add_digits(s):
    val = 0
    for c in s:
        val += int(c)
    return val

def mul(x, y):
    tot = 0
    for i in range(y):
        tot += x
    return tot


def fact_iter(n):
    prod = 1
    for i in range(1, n+1):
        prod *= i
    return prod

def fact_recur(n):
    """ assume n >= 0 """
    if n <= 1: 
        return 1
    else: 
        return n*fact_recur(n - 1)
        
def linear_search(L, e):
    found = False
    for i in range(len(L)):
        if e == L[i]:
            found = True
    return found

test_list = [1, 3, 4, 5, 9, 18, 27]
#print(linear_search(test_list, 9))

def search(L, e):
    for i in L:
        if i == e:
            return True
        if i > e:
            return False
    return False

# test_list = [1, 3, 4, 5, 9, 18, 27]
#print(search(test_list, 9))

def fib_iter(n):
    if n == 0:
        return 0
    elif n == 1:
        return 1
    else:
        fib_i = 0
        fib_ii = 1
        for i in range(n-1):
            tmp = fib_i
            fib_i = fib_ii
            fib_ii = tmp + fib_ii
        return fib_ii 

##################
## Example: Polynomial Complexity
##################
def g(n):
   """ assume n >= 0 """
   x = 0
   for i in range(n):
      for j in range(n):
         x += 1
   return x

def is_subset(L1, L2):
    for e1 in L1:
        matched = False
        for e2 in L2:
            if e1 == e2:
                matched = True
                break
        if not matched:
            return False
    return True

#test_set = [1, 2, 3, 4, 5]
#test_set1 = [1, 5, 3]
#test_set2 = [1, 6]
#print(is_subset(test_set1, test_set))

def intersect(L1, L2):
    tmp = []
    for e1 in L1:
        for e2 in L2:
            if e1 == e2:
                tmp.append(e1)
    unique = []
    for e in tmp:
        if not(e in unique):
            unique.append(e)
    return unique

#print(intersect(test_set1, test_set2))

##################
## Example: Exponential Complexity
##################
    
##### binary numbers up to n bits
    
def all_binary_numbers(N):
    def helper (prefix, N):
        if N == 0:
            return [prefix]
        return helper(prefix+'0', N-1) + helper(prefix+'1', N-1)                        
    return helper('', N)

#for N in range(1, 6):
#     print (all_binary_numbers(N))
     
#L_N = [5, 10, 15, 20, 25]
#
#times = []
#for N in L_N:
#    t = time.perf_counter()
#    abn = all_binary_numbers(N)
#    t = time.perf_counter()-t
#    print ('binary numbers of ', N, 'digits took ', t, 's')
#    times.append(t)
#    
#plt.plot(L_N, times, 'bo')
#plt.xlabel('Number of Binary Digits')
#plt.ylabel('Time in Seconds')
#plt.title('Measured Time of all_binary_numbers')
#plt.semilogy()

def bisect_search1(L, e):
    if L == []:
        return False
    elif len(L) == 1:
        return L[0] == e
    else:
        half = len(L)//2
        if L[half] > e:
            return bisect_search1( L[:half], e)
        else:
            return bisect_search1( L[half:], e)

#L = [1,2,3,4,5,6,7,8]
#print(bisect_search1(L, 3.5))


def is_sorted(L):
    result = True
    for i in range(len(L) - 1):
        if L[i] > L[i+1]:
            result = False
    return result
    
def stupid_sort(L):
    while not is_sorted(L):
        random.shuffle(L)
        print(L)

# L = [2,3,1,9,0]
# stupid_sort(L)
# print(L)

def selection_sort(L):
    suffix = 0
    while suffix != len(L):
        for i in range(suffix, len(L)):
            if L[i] < L[suffix]:
                L[suffix], L[i] = L[i], L[suffix]
        suffix += 1

# L = [2,3,1,9,0]        
# selection_sort(L)
# print(L)

def merge(left, right):
    result = []
    i,j = 0, 0
    while i < len(left) and j < len(right):
        if left[i] < right[j]:
            result.append(left[i])
            i += 1
        else:
            result.append(right[j])
            j += 1
    while (i < len(left)):
        result.append(left[i])
        i += 1
    while (j < len(right)):
        result.append(right[j])
        j += 1
    return result

def merge_sort(L):
    if len(L) < 2:
        return L[:]
    else:
        middle = len(L)//2
        left = merge_sort(L[:middle])
        right = merge_sort(L[middle:])
        return merge(left, right)

